SystemOrganization addCategory: #MetaInterpreter!

TestCase subclass: #TestInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:13'!
passAsArg: t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 18:23'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter onMessageReceivedForObject: TestObject do: [:reifications :override | 
		(reifications value: #message) 
			selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter onArgumentForObject: t2 do: [:reifications :override |
			(reifications value: #receiver) = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter onMessageSendFromObject: t1 do: [:reifications :override | 
		(reifications value: #receiver) = t2 ifTrue: [self halt].
	].

	Interpreter onMessageSendFromObject: t2 do: [:reifications :override | 
		(reifications value: #receiver) = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.
	Interpreter enableMetaBehaviorFor: t2.

	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
	Interpreter disableMetaBehaviorFor: t2.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 15:27'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onArgumentForObject: t do: [:trigger :context :method  :sender :receiver :message :override |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 15:28'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onIvReadForObject: t do: [:trigger :context :index :method :override | 
			ivReadCounter := ivReadCounter + 1.			
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 15:28'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:trigger :context :method :sender :message :override | 
			receiveCounter := receiveCounter + 1.		
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 15:28'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter onMessageSendFromObject: t do: [:trigger :context  :method :receiver :message :override | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 15:29'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter onIvWriteForObject: t do: [:trigger :context :index :value :method :override | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 17:22'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:trigger :context :method :sender :message :override | 
			receiveCounter := receiveCounter + 1.		
		].
	
		Interpreter onMessageSendFromObject: t do: [:trigger :context  :method :receiver :message :override | 
			sendCounter := sendCounter + 1.	
		].
	
		Interpreter onIvReadForObject: t do: [:trigger :context :index :method :override | 
			ivReadCounter := ivReadCounter + 1.			
		].

		Interpreter onIvWriteForObject: t do: [:trigger :context :index :value :method :override | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter onArgumentForObject: t do: [:trigger :context :method  :sender :receiver :message :override |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 19:37'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter onMessageReceivedForObject: t do: [:trigger :context :method :sender :message :override | 
		self halt.
		receiveCounter := receiveCounter + 1.		
	].
	
	Interpreter onMessageSendFromObject: t do: [:trigger :context  :method :receiver :message :override | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter onIvReadForObject: t do: [:trigger :context :index :method :override | 
		self halt.
		ivReadCounter := ivReadCounter + 1.			
	].

	Interpreter onIvWriteForObject: t do: [:trigger :context :index :value :method :override | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.		
	].

	Interpreter onArgumentForObject: t do: [:trigger :context :method  :sender :receiver :message :override |
		self halt.
		argCounter := argCounter + 1
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 17:45'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:trigger :context :method :sender :message :override | 
			receiveCounter := receiveCounter + 1.		
		].
	
		Interpreter onMessageSendFromObject: t do: [:trigger :context :method :receiver :message :override | 
			sendCounter := sendCounter + 1.	
		].
	
		Interpreter onIvReadForObject: t do: [:trigger :context :index :method :override | 
			ivReadCounter := ivReadCounter + 1.
			override true. "this callback will override default behavior"
			Interpreter enableMetaBehaviorFor: trigger. "in meta by default the meta-jump is disabled
			since we override we explicitely enable it back"
			666. "return value of iv read"	
		].

		Interpreter onIvWriteForObject: t do: [:trigger :context :index :value :method :override | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter onArgumentForObject: t do: [:trigger :context :method :sender :receiver :message :override |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

Object subclass: #BenchmarkObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!BenchmarkObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
test
	| a b c |
	a := 1.
	b := 2.
	c := a + b.
	^ c
! !

Object subclass: #ExplicitMetaObject
	instanceVariableNames: 'implicitMetaObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!ExplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/31/2013 12:56'!
implicitMetaObject
	^ implicitMetaObject! !

!ExplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/31/2013 12:56'!
implicitMetaObject: anObject
	implicitMetaObject := anObject! !

ExplicitMetaObject subclass: #Reflectogram
	instanceVariableNames: 'targetObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:19'!
newOn: aTargetObject

	self new 
		implicitMetaObject: LowLevelImplicitMetaObject new;
		targetObject: aTargetObject  ! !

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:35'!
of: aTargetObject

	^ Interpreter baseMetaDict at: aTargetObject  ifAbsent: [
		Interpreter baseMetaDict at: aTargetObject  put: (Reflectogram newOn: aTargetObject) 	
	].! !

!Reflectogram methodsFor: 'meta-jump-control' stamp: 'NickPapoulias 6/1/2013 18:25'!
disable

	Interpreter disableMetaBehaviorFor: targetObject ! !

!Reflectogram methodsFor: 'meta-jump-control' stamp: 'NickPapoulias 6/1/2013 18:25'!
enable

	Interpreter enableMetaBehaviorFor: targetObject ! !

!Reflectogram methodsFor: 'meta-jump-control' stamp: 'NickPapoulias 6/1/2013 16:51'!
isEnabled

	^ Interpreter isMetaBehaviorEnabledFor: targetObject ! !

!Reflectogram methodsFor: 'meta-context-control' stamp: 'NickPapoulias 6/1/2013 16:47'!
metaLevel

	Processor  activeProcess metaLevel ! !

!Reflectogram methodsFor: 'temporal-selection' stamp: 'NickPapoulias 6/1/2013 16:48'!
on: anEvent do: aCallback

	anEvent registerCallback: aCallback to: implicitMetaObject ! !

!Reflectogram methodsFor: 'default-meta-action' stamp: 'NickPapoulias 6/1/2013 18:55'!
performDefaultActionFor: anEvent with: reifications 

	^ anEvent performDefaultActionFor: targetObject with: reifications.
	! !

!Reflectogram methodsFor: 'accessing' stamp: 'NickPapoulias 6/1/2013 16:51'!
targetObject

	^ targetObject! !

!Reflectogram methodsFor: 'accessing' stamp: 'NickPapoulias 6/1/2013 16:51'!
targetObject: anObject

	targetObject := anObject! !

Object subclass: #Interpreter
	instanceVariableNames: ''
	classVariableNames: 'BaseMetaDict'
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/8/2013 15:04'!
baseMetaDict

	^ BaseMetaDict! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:12'!
disableMetaBehaviorFor: anObject

	"unbind"
	
	self primitiveSetExperimentalBitOf: anObject to: false. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:11'!
enableMetaBehaviorFor: anObject

	"bind"
	
	self primitiveSetExperimentalBitOf: anObject to: true. ! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/8/2013 14:28'!
initialize

	super initialize.
	BaseMetaDict := IdentityDictionary new.
	Smalltalk specialObjectsArray becomeForward: self newSpecialObjectsArray.! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:13'!
isMetaBehaviorEnabledFor: anObject


	^ self primitiveGetExperimentalBitOf: anObject! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/29/2013 14:17'!
newSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	
	| newArray |
	newArray := Array new: 76.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (Smalltalk globals associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: nil. "Was BlockContext"
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (Smalltalk specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 nil 0
							nil 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (Smalltalk specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: Smalltalk compactClassesArray.
	newArray at: 30 put: (Smalltalk specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (Smalltalk specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: 0@0.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (Smalltalk specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (Smalltalk specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (Smalltalk specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((Smalltalk specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (Smalltalk at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (Smalltalk at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (Smalltalk at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (Smalltalk at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (Smalltalk at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (Smalltalk at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (Smalltalk at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (Smalltalk at: #WeakFinalizationList ifAbsent: []).
	newArray at: 59 put: #metaReceived:.
	newArray at: 60 put: #metaRef.
	newArray at: 61 put: #metaReadAt:.
	newArray at: 62 put: #metaWriteAt:Put:.
	newArray at: 63 put: #metaArgIn:for:.
	newArray at: 64 put: Interpreter.
	newArray at: 65 put: #metaSend:from:.
	newArray at: 66 put: Interpreter baseMetaDict.
	"optimization"
	newArray at: 67 put: Interpreter class>>#object:metaReceived:.
	newArray at: 68 put: Interpreter class>>#object:metaSend:withReceiver:.
	newArray at: 69 put: Interpreter class>>#object:metaReadAt:.
	newArray at: 70 put: Interpreter class>>#object:metaWriteAt:Put:.
	newArray at: 71 put: Interpreter class>>#object:metaArgIn:withReceiver:.
	
	newArray at: 72 put: #object:metaReceived:.
	newArray at: 73 put: #object:metaSend:withReceiver:.
	newArray at: 74 put: #object:metaReadAt:.
	newArray at: 75 put: #object:metaWriteAt:Put:.
	newArray at: 76 put: #object:metaArgIn:withReceiver:.
	^newArray! !

!Interpreter class methodsFor: 'introspection' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex

	<primitive: 73>
	! !

!Interpreter class methodsFor: 'modification' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex put: anotherObject

	<primitive: 74>
	! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/1/2013 18:39'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| implicitMetaObject reifications reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr."

 	reifications :=	[:value |
		value = #trigger ifTrue: [ anObject ].
		value = #context ifTrue: [ thisContext home sender ].
		value = #method ifTrue: [ thisContext home sender method ].
		value = #sender ifTrue: [ thisContext home sender receiver ].
		value = #receiver ifTrue: [ rcvr ].
		value = #message ifTrue: [ aMessage ].
		value = #sourceRange ifTrue: [ thisContext home sender debuggerMap 
						rangeForPC: thisContext home sender pc
						contextIsActiveContext: false ].
	]. "lazy evaluation of reifications"

	reflectogram := Reflectogram of: anObject.

	implicitMetaObject := reflectogram implicitMetaObject.

	^ implicitMetaObject argCallback valueWithArguments: { reifications . reflectogram}.
	! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/1/2013 18:57'!
object: anObject metaReadAt: anIndex

	| implicitMetaObject reifications reflectogram |
	"Transcript show: thisContext; cr."

	reifications :=	[:value |
		value = #trigger ifTrue: [ anObject ].
		value = #context ifTrue: [ thisContext home sender ].
		value = #method ifTrue: [ thisContext home sender method ].
		value = #index ifTrue: [ anIndex ].
		value = #sourceRange ifTrue: [ thisContext home sender debuggerMap 
						rangeForPC: thisContext home sender pc
						contextIsActiveContext: false ].
	]. "lazy evaluation of reifications"
	
	reflectogram := Reflectogram of: anObject.

	implicitMetaObject := reflectogram implicitMetaObject.

	^ implicitMetaObject argCallback valueWithArguments: { reifications . reflectogram}.! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/1/2013 18:58'!
object: anObject metaReceived: aMessage

	| implicitMetaObject reifications reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr."
	
	reifications :=	[:value |
		value = #trigger ifTrue: [ anObject ].
		value = #context ifTrue: [ thisContext home sender ].
		value = #method ifTrue: [ thisContext home sender method ].
		value = #sender ifTrue: [ thisContext home sender receiver ].
		value = #message ifTrue: [ aMessage ].
		value = #sourceRange ifTrue: [ thisContext home sender debuggerMap 
						rangeForPC: thisContext home sender pc
						contextIsActiveContext: false ].
	]. "lazy evaluation of reifications"

	
	reflectogram := Reflectogram of: anObject.

	implicitMetaObject := reflectogram implicitMetaObject.

	^ implicitMetaObject argCallback valueWithArguments: { reifications . reflectogram}.! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/1/2013 18:58'!
object: anObject metaSend: aMessage withReceiver: anotherObject

	| implicitMetaObject reifications reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr."

	reifications :=	[:value |
		value = #trigger ifTrue: [ anObject ].
		value = #context ifTrue: [ thisContext home sender ].
		value = #method ifTrue: [ thisContext home sender method ].
		value = #receiver ifTrue: [ anotherObject  ].
		value = #message ifTrue: [ aMessage ].
		value = #sourceRange ifTrue: [ thisContext home sender debuggerMap 
						rangeForPC: thisContext home sender pc
						contextIsActiveContext: false ].
	]. "lazy evaluation of reifications"

						
	reflectogram := Reflectogram of: anObject.

	implicitMetaObject := reflectogram implicitMetaObject.

	^ implicitMetaObject argCallback valueWithArguments: { reifications . reflectogram}.! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/1/2013 18:58'!
object: anObject metaWriteAt: anIndex Put: anotherObject

	| implicitMetaObject reifications reflectogram |
	"Transcript show: thisContext; cr."
	
	reifications :=	[:value |
		value = #trigger ifTrue: [ anObject ].
		value = #context ifTrue: [ thisContext home sender ].
		value = #index ifTrue: [ anIndex ].
		value = #value ifTrue: [ anotherObject ].
		value = #method ifTrue: [ thisContext home sender method ].
		value = #sourceRange ifTrue: [ thisContext home sender debuggerMap 
						rangeForPC: thisContext home sender pc
						contextIsActiveContext: false ].
	]. "lazy evaluation of reifications"
						
	reflectogram := Reflectogram of: anObject.

	implicitMetaObject := reflectogram implicitMetaObject.

	^ implicitMetaObject argCallback valueWithArguments: { reifications . reflectogram}.! !

!Interpreter class methodsFor: 'execution' stamp: 'NickPapoulias 5/5/2013 18:44'!
object: anObject perform: aSelector withArguments: args 

	<primitive: 84>! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onArgumentForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject argCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvReadForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject readCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvWriteForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject writeCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageReceivedForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject receiveCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageSendFromObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject sendCallback: aCallback.
! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:47'!
primitiveGetExperimentalBitOf: anObject

	<primitive: 150>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:22'!
primitiveSetExperimentalBitOf: anObject to: aBoolean

	<primitive: 151>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/9/2013 19:22'!
removeMetaBehaviorOf: anObject

	"unbind and remove implicit meta-object"
	self disableMetaBehaviorFor: anObject.
	BaseMetaDict removeKey: anObject.! !

Object subclass: #LowLevelImplicitMetaObject
	instanceVariableNames: 'receiveCallback sendCallback readCallback writeCallback argCallback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LowLevelImplicitMetaObject class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/7/2013 17:57'!
new

	^ self basicNew initialize! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback
	^ argCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback: anObject
	argCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/29/2013 18:20'!
initialize 

	super initialize.

	self receiveCallback: [:reifications :override | ].

	self sendCallback: [:reifications :override | ].

	self readCallback: [:reifications :override | ].

	self writeCallback: [:reifications :override | ].
	
	self	 argCallback: [:reifications :override | ].
		! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback
	^ readCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback: anObject
	readCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback
	^ receiveCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback: anObject
	receiveCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback
	^ sendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback: anObject
	sendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback
	^ writeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback: anObject
	writeCallback := anObject! !

Object subclass: #MetaBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:36'!
benchmark: aBlock times: anInteger
	| runs average deviation |

	runs := [ 
		(1 to: anInteger) 
			collect: [ :each | aBlock timeToRun ] ]
		valueUnpreemptively.
	average := runs average asFloat.
	deviation := ((runs 
		detectSum: [ :each | (each - average) * (each - average) ]) 
			/ runs size) sqrt.
	^ Array with: average with: deviation! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
message
	"self new message" 
	
	|  object  |
	object := BenchmarkObject new.
	^ self benchmark: [ 1 to: 1000000 do: [ :each | object test ] ] times: 100.! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/29/2013 18:34'!
messageMetaSend
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter onMessageReceivedForObject: object do: [:reifications :override | override false].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit
	^ self primitiveGetExperimentalBit ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit: aBoolean
	self primitiveSetExperimentalBitTo: aBoolean! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/8/2013 15:34'!
identityHash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	This method must not be overridden, except by SmallInteger.
	
	Do not override."

	^ (Interpreter object: self perform: #basicIdentityHash withArguments: {}) bitShift: 18! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
isExperimentalBitOn
	^ self experimentalBit! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaArgIn: aMessage for: anObject

	^ Interpreter object: anObject metaArgIn: aMessage withReceiver: self! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReadAt: anIndex

	^ Interpreter object: self metaReadAt: anIndex! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReceived: aMessage

	^ Interpreter object: self metaReceived: aMessage ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:39'!
metaRef

	self halt! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaSend: aMessage from: anObject

	^ Interpreter object: anObject metaSend: aMessage withReceiver: self.! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaWriteAt: anIndex Put: anotherObject

	^ Interpreter object: self metaWriteAt: anIndex Put: anotherObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveGetExperimentalBit

	<primitive: 150>
	self primitiveFailed! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveSetExperimentalBitTo: aBoolean

	<primitive: 151>
	self primitiveFailed! !

Object subclass: #Operation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #AsArgument
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!AsArgument class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:56'!
performDefaultActionFor: targetObject with: reifications

	^ Interpreter object: (reifications value: #receiver) 
				perform: (reifications value: #message) selector 
				withArguments: (reifications value: #message) arguments.	! !

!AsArgument class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:05'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject argCallback: aCallback ! !

Operation subclass: #AsReturnValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #CompositeMetaEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeMetaEvent subclass: #ObjectCreation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeMetaEvent subclass: #ObjectInteraction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #IvRead
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvRead class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:56'!
performDefaultActionFor: targetObject with: reifications

	^ Interpreter object: targetObject at: (reifications value: #index).! !

!IvRead class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:05'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject readCallback: aCallback ! !

Operation subclass: #IvWrite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvWrite class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:56'!
performDefaultActionFor: targetObject with: reifications

	^ Interpreter object: targetObject at: (reifications value: #index) put: (reifications value: #anotherObject).! !

!IvWrite class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:06'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject writeCallback: aCallback ! !

Operation subclass: #MessageReceived
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageReceived class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:56'!
performDefaultActionFor: targetObject with: reifications

	^ Interpreter object: targetObject perform: (reifications value: #message) selector withArguments: (reifications value: #message) arguments.! !

!MessageReceived class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:11'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject receiveCallback: aCallback ! !

Operation subclass: #MessageSent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageSent class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 18:56'!
performDefaultActionFor: targetObject with: reifications

	^ Interpreter object: (reifications value: #receiver) perform: (reifications value: #message) selector withArguments: (reifications value: #message) arguments.! !

!MessageSent class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:12'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject sendCallback: aCallback ! !

!Operation class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/1/2013 16:34'!
performDefaultActionFor: targetObject with: reifications

	self subclassResponsibility ! !

!Operation class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/31/2013 14:04'!
registerCallback: aCallback to: anImplicitMetaObject

	self subclassResponsibility ! !

Operation subclass: #Referenced
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #Stored
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Object subclass: #TestObject
	instanceVariableNames: 'x'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 16:04'!
testInteraction: anObject 

	anObject class.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 15:58'!
testPassingArgument: anObject ! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/9/2013 19:38'!
testSenderHook

	Transcript show: 'this message should be caught'; cr.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:01'!
x

	^ x! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:02'!
x: anObject 

	x := anObject ! !
Interpreter initialize!
